import unittest
import datetime
import json
import os

from asyncwhois.parse import BaseParser, convert_whodap_keys


class TestWhoIsParserMethods(unittest.TestCase):
    def test_convert_whodap_keys(self):
        with open(
            os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "samples/com_dict_asyncwhois.json",
            )
        ) as o:
            asyncwhois_dict = json.loads(o.read())

        with open(
            os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "samples/com_dict_whodap.json",
            )
        ) as o:
            whodap_dict = json.loads(o.read())

        whodap_keys_after = convert_whodap_keys(whodap_dict).keys()
        asyncwhois_keys = asyncwhois_dict.keys()
        for whodap_key in whodap_keys_after:
            assert (
                whodap_key in asyncwhois_keys
            ), f"{whodap_key} not in {asyncwhois_keys}"

        for asyncwhois_key in asyncwhois_dict.keys():
            assert (
                asyncwhois_key in whodap_keys_after
            ), f"{asyncwhois_key} not in {whodap_keys_after}"

    def test_parse_dates(self):
        date_strings = [
            "11-aug-2020",
            "11-August-2020",
            "11-09-2020",
            "2020-09-20",
            "2020.09.20",
            "2020/09/20",
            "2020. 09. 20.",
            "2020.09.20 11:11:11",
            "August 11 2020",
            "20200920",
        ]

        for date_string in date_strings:
            formatted_date = BaseParser._parse_date(date_string)
            self.assertIsInstance(formatted_date, datetime.datetime)

    def test_find_match(self):
        test_blob = """
        Domain name: google.com
        Name server: ns1.google.com
        Name server: ns2.google.com
        Status: ok
        """
        domain = BaseParser().find_match(r"domain name: *(.+)", test_blob)
        self.assertEqual(domain, "google.com")
        name_servers = BaseParser().find_match(
            r"name server: *(.+)", test_blob, many=True
        )
        self.assertEqual(len(name_servers), 2)
        status = BaseParser().find_match(r"status: *(.+)", test_blob, many=True)
        self.assertEqual(len(status), 1)

    def test_find_multiline_match(self):
        test_blob = """
        Domain name: google.com
        
        Domain nameservers:
           ns1.googledomains.com
           ns2.googledomains.com
           ns3.googledomains.com
           ns4.googledomains.com

        Registrar: someone
        """
        name_servers = BaseParser().find_multiline_match(
            r"Domain nameservers:\n", test_blob
        )
        self.assertEqual(len(name_servers), 4)



import sys

import asyncwhois
from asyncwhois.errors import NotFoundError

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase

    class TestLookupNotFound(IsolatedAsyncioTestCase):

        async def test_not_found_aio(self):
            domain = 'some-non-existent-domain123.com'
            with self.assertRaises(NotFoundError):
                await asyncwhois.aio_whois_domain(domain)

        def test_not_found(self):
            domain = 'some-non-existent-domain123.com'
            with self.assertRaises(NotFoundError):
                asyncwhois.whois_domain(domain)



from ipaddress import IPv4Address, IPv6Address
from typing import Any, Optional, Union

from .pywhois import DomainLookup, NumberLookup, ASNLookup

__all__ = [
    "aio_whois_domain",
    "aio_whois_ipv4",
    "aio_whois_ipv6",
    "aio_rdap_domain",
    "aio_rdap_ipv4",
    "aio_rdap_ipv6",
    "aio_rdap_asn",
    "rdap_domain",
    "rdap_ipv4",
    "rdap_ipv6",
    "rdap_asn",
    "whois_domain",
    "whois_ipv4",
    "whois_ipv6",
]
__version__ = "1.0.9"


def whois_domain(
    domain: str,
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> DomainLookup:
    """
    Performs domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of DomainLookup
    """
    result = DomainLookup.whois_domain(domain, authoritative_only, proxy_url, timeout)
    return result


async def aio_whois_domain(
    domain: str,
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> DomainLookup:
    """
    Performs asynchronous domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of DomainLookup
    """
    result = await DomainLookup.aio_whois_domain(
        domain, authoritative_only, proxy_url, timeout
    )
    return result


def rdap_domain(domain: str, httpx_client: Optional[Any] = None) -> DomainLookup:
    """
    Performs an RDAP query for the given domain.
    Finds the authoritative RDAP server and parses the response from that server.

    :param domain: Any domain name or URL
        (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of DomainLookup
    """
    result = DomainLookup.rdap_domain(domain, httpx_client)
    return result


async def aio_rdap_domain(
    domain: str, httpx_client: Optional[Any] = None
) -> DomainLookup:
    """
    Performs an async RDAP query for the given domain name.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of DomainLookup
    """
    result = await DomainLookup.aio_rdap_domain(domain, httpx_client)
    return result


def whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of DomainLookup
    """
    result = NumberLookup.whois_ipv4(ipv4, authoritative_only, proxy_url, timeout)
    return result


async def aio_whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    result = await NumberLookup.aio_whois_ipv4(
        ipv4, authoritative_only, proxy_url, timeout
    )
    return result


def rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv4 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    result = NumberLookup.rdap_ipv4(ipv4, httpx_client)
    return result


async def aio_rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    result = await NumberLookup.aio_rdap_ipv4(ipv4, httpx_client)
    return result


def whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    result = NumberLookup.whois_ipv6(ipv6, authoritative_only, proxy_url, timeout)
    return result


async def aio_whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    result = await NumberLookup.aio_whois_ipv6(
        ipv6, authoritative_only, proxy_url, timeout
    )
    return result


def rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    result = NumberLookup.rdap_ipv6(ipv6, httpx_client)
    return result


async def aio_rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    result = await NumberLookup.aio_rdap_ipv6(ipv6, httpx_client)
    return result


def rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of ASNLookup
    """
    result = ASNLookup.rdap_asn(asn, httpx_client)
    return result


async def aio_rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an async RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of ASNLookup
    """
    result = await ASNLookup.aio_rdap_asn(asn, httpx_client)
    return result



import unittest
import unittest.mock as mock

import asyncwhois.query
from asyncwhois.errors import QueryError


class TestWhoIsQuery(unittest.TestCase):

    @mock.patch('asyncwhois.query.socket')
    def test_whois_query_create_connection(self, mock_socket_lib):
        # test connect
        test_address_tuple_param = ('0.0.0.0', 69)
        test_timeout_param = 10
        with asyncwhois.query.Query()._create_connection(test_address_tuple_param) as _:
            ...
        mock_socket_lib.create_connection.assert_called()
        mock_socket_lib.create_connection.assert_called_with(test_address_tuple_param,
                                                             test_timeout_param)

    @mock.patch('asyncwhois.query.socket.socket')
    def test_whois_query_send_and_recv(self, mock_socket_instance):
        test_data_send_string = "a-domain-to-send"
        test_data_recv_bytes = b""  # empty so _send_and_recv does not infinite loop
        mock_socket_instance.recv.return_value = test_data_recv_bytes
        asyncwhois.query.Query._send_and_recv(mock_socket_instance, test_data_send_string)
        mock_socket_instance.recv.assert_called()
        mock_socket_instance.sendall.assert_called()
        mock_socket_instance.sendall.assert_called_with(test_data_send_string.encode())



import ipaddress
from typing import Union, Dict, Any, Optional

import tldextract
import whodap

from .parse import convert_whodap_keys
from .parse_rir import NumberParser
from .parse_tld import DomainParser
from .query import DomainQuery, NumberQuery
from .servers import CountryCodeTLD, GenericTLD, SponsoredTLD, IPv4Allocations


class Lookup:
    def __init__(self):
        self._query = None
        self._parser = None

    @property
    def parser_output(self) -> Dict[str, Any]:
        if isinstance(self._parser, dict):
            return self._parser
        elif isinstance(self._parser, (DomainParser, NumberParser)):
            return self._parser.parser_output

    @property
    def query_output(self) -> str:
        if isinstance(self._query, (DomainQuery, NumberQuery)):
            if not self._query.authoritative_only:
                # return entire query chain
                return self._query.query_chain
            else:
                # return only the authoritative response
                return self._query.query_output
        else:
            return self._query


class DomainLookup(Lookup):
    def __init__(self):
        super().__init__()
        self._tld_extract = None

    @staticmethod
    def _get_server_name(tld: str) -> Union[str, None]:
        tld_converted = tld.upper().replace("-", "_")
        for servers in [CountryCodeTLD, GenericTLD, SponsoredTLD]:
            if hasattr(servers, tld_converted):
                server = getattr(servers, tld_converted)
                return server
        return None

    def _get_top_level_domain(self, domain: str):
        self._tld_extract = tldextract.extract(domain)
        top_level_domain = self._tld_extract.suffix
        if "." in top_level_domain:
            top_level_domain = top_level_domain.split(".")[-1]
        return top_level_domain

    @property
    def tld_extract_result(self) -> tldextract.tldextract.ExtractResult:
        return self._tld_extract

    @classmethod
    def whois_domain(
        cls, domain: str, authoritative_only: bool, proxy_url: str, timeout: int
    ):
        _self = cls()
        # get TLD
        top_level_domain = _self._get_top_level_domain(domain)
        # initialize parser based on given TLD
        parser = DomainParser(top_level_domain)
        # get the WHOIS server associated with this TLD
        server = _self._get_server_name(top_level_domain)
        # submit the WHOIS query to the server
        registered_domain = _self.tld_extract_result.registered_domain
        query = DomainQuery.new(
            registered_domain, server, authoritative_only, proxy_url, timeout
        )
        # parse the raw text output from the WHOIS server
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    async def aio_whois_domain(
        cls, domain: str, authoritative_only: bool, proxy_url: str, timeout: int
    ):
        _self = cls()
        # get TLD
        top_level_domain = _self._get_top_level_domain(domain)
        # initialize parser based on given TLD
        parser = DomainParser(top_level_domain)
        # get the WHOIS server associated with this TLD
        server = _self._get_server_name(top_level_domain)
        # submit the WHOIS query to the server
        registered_domain = _self.tld_extract_result.registered_domain
        query = await DomainQuery.new_aio(
            registered_domain, server, authoritative_only, proxy_url, timeout
        )
        # parse the raw text output from the WHOIS server
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    def rdap_domain(cls, domain: str, httpx_client: Any = None):
        """
        Performs an RDAP query using the `whodap` project.
        Stores the resulting RDAP output into "query_output" and a WHOIS friendly
        key:value pair dictionary into "parser_output".

        :param domain: the domain or URL to search (e.g. 'google.com' or 'https://www.google.com')
        :param httpx_client: the underlying httpx client to pass to `whodap.lookup_domain`
        :return: instance of DomainLookup
        """
        _self = cls()
        top_level_domain = _self._get_top_level_domain(domain)
        response = whodap.lookup_domain(
            domain=_self.tld_extract_result.domain,
            tld=top_level_domain,
            httpx_client=httpx_client,
        )
        _self._query = response.to_dict()
        whois_dict = response.to_whois_dict()
        _self._parser = convert_whodap_keys(whois_dict)
        return _self

    @classmethod
    async def aio_rdap_domain(cls, domain: str, httpx_client: Any = None):
        """
        Performs an async RDAP query using the `whodap` project.
        Stores the resulting RDAP output into "query_output" and a WHOIS friendly
        key:value pair dictionary into "parser_output".

        :param domain: the domain or URL to search (e.g. 'google.com' or 'https://www.google.com')
        :param httpx_client: the underlying httpx client to pass to `whodap.aio_lookup_domain`
        :return: instance of DomainLookup
        """
        _self = cls()
        top_level_domain = _self._get_top_level_domain(domain)
        response = await whodap.aio_lookup_domain(
            domain=_self.tld_extract_result.domain,
            tld=top_level_domain,
            httpx_client=httpx_client,
        )
        _self._query = response.to_dict()
        whois_dict = response.to_whois_dict()
        _self._parser = convert_whodap_keys(whois_dict)
        return _self


class NumberLookup(Lookup):
    def __init__(self):
        super().__init__()
        self._ip = None

    @property
    def ipv6(self) -> Optional[ipaddress.IPv6Address]:
        if isinstance(self._ip, ipaddress.IPv6Address):
            return self._ip

    @property
    def ipv4(self) -> Optional[ipaddress.IPv4Address]:
        if isinstance(self._ip, ipaddress.IPv4Address):
            return self._ip

    @staticmethod
    def convert_to_ip(ip: str):
        try:
            ip_obj = ipaddress.ip_address(ip)
            return ip_obj
        except ipaddress.AddressValueError:
            raise

    @classmethod
    def whois_ipv4(
        cls,
        ipv4: Union[str, ipaddress.IPv4Address],
        authoritative_only: bool,
        proxy_url: str,
        timeout: int,
    ):
        _self = cls()
        if not isinstance(ipv4, ipaddress.IPv4Address):
            _self._ip = _self.convert_to_ip(ipv4)
        else:
            _self._ip = ipv4
        _, server = IPv4Allocations().get_servers(_self._ip)
        ip_string = str(_self._ip)
        # run the query
        query = NumberQuery.new(
            ip_string, server, authoritative_only, proxy_url, timeout
        )
        # parse the raw text output from the WHOIS server
        parser = NumberParser(query.authoritative_server)
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    async def aio_whois_ipv4(
        cls,
        ipv4: Union[str, ipaddress.IPv4Address],
        authoritative_only: bool,
        proxy_url: str,
        timeout: int,
    ):
        _self = cls()
        if not isinstance(ipv4, ipaddress.IPv4Address):
            _self._ip = _self.convert_to_ip(ipv4)
        else:
            _self._ip = ipv4
        _, server = IPv4Allocations().get_servers(_self._ip)
        ip_string = str(_self._ip)
        # run the query
        query = await NumberQuery.new_aio(
            ip_string, server, authoritative_only, proxy_url, timeout
        )
        # parse the raw text output from the WHOIS server
        parser = NumberParser(query.authoritative_server)
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    def rdap_ipv4(
        cls, ipv4: Union[str, ipaddress.IPv4Address], httpx_client: Any = None
    ):
        _self = cls()
        response = whodap.lookup_ipv4(ipv4, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None
        return _self

    @classmethod
    async def aio_rdap_ipv4(
        cls, ipv4: Union[str, ipaddress.IPv4Address], httpx_client: Any = None
    ):
        _self = cls()
        response = await whodap.aio_lookup_ipv4(ipv4, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None
        return _self

    @classmethod
    def whois_ipv6(
        cls,
        ipv6: Union[str, ipaddress.IPv6Address],
        authoritative_only: bool,
        proxy_url: str,
        timeout: int,
    ):
        _self = cls()
        if not isinstance(ipv6, ipaddress.IPv6Address):
            ipv6 = _self.convert_to_ip(ipv6)
        ip_string = str(ipv6)
        # run the query
        query = NumberQuery.new(ip_string, None, authoritative_only, proxy_url, timeout)
        # parse the raw text output from the WHOIS server
        parser = NumberParser(query.authoritative_server)
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    async def aio_whois_ipv6(
        cls,
        ipv6: Union[str, ipaddress.IPv6Address],
        authoritative_only: bool,
        proxy_url: str,
        timeout: int,
    ):
        _self = cls()
        if not isinstance(ipv6, ipaddress.IPv4Address):
            _self._ip = _self.convert_to_ip(ipv6)
        ip_string = str(_self._ip)
        # run the query
        query = await NumberQuery.new_aio(
            ip_string, None, authoritative_only, proxy_url, timeout
        )
        # parse the raw text output from the WHOIS server
        parser = NumberParser(query.authoritative_server)
        parser.parse(query.query_output)
        _self._query = query
        _self._parser = parser
        return _self

    @classmethod
    def rdap_ipv6(
        cls, ipv6: Union[str, ipaddress.IPv6Address], httpx_client: Any = None
    ):
        _self = cls()
        response = whodap.lookup_ipv6(ipv6, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None
        return _self

    @classmethod
    async def aio_rdap_ipv6(
        cls, ipv6: Union[str, ipaddress.IPv6Address], httpx_client: Any = None
    ):
        _self = cls()
        response = await whodap.aio_lookup_ipv6(ipv6, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None
        return _self


class ASNLookup(Lookup):
    def __init__(self):
        super().__init__()
        self.asn: int = None

    @classmethod
    def rdap_asn(cls, asn: int, httpx_client: Any = None):
        _self = cls()
        response = whodap.lookup_asn(asn, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None  # parser support not implemented
        return _self

    @classmethod
    async def aio_rdap_asn(cls, asn: int, httpx_client: Any = None):
        _self = cls()
        response = await whodap.aio_lookup_asn(asn, httpx_client)
        _self._query = response.to_dict()
        _self._parser = None
        return _self



from datetime import datetime
import re
from typing import Dict, Any, Union

from .parse import BaseParser, TLDBaseKeys
from .errors import NotFoundError


class TLDParser(BaseParser):
    base_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name: *(.+)",
        TLDBaseKeys.CREATED: r"Creation Date: *(.+)",
        TLDBaseKeys.UPDATED: r"Updated Date: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expir\w+\sDate: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar: *(.+)",
        TLDBaseKeys.REGISTRAR_IANA_ID: r"Registrar IANA ID: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Registrar URL: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"Registrar Abuse Contact Email: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"Registrar Abuse Contact Phone: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Name: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Registrant Organization: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Registrant Street: *(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"Registrant City: *(.+)",
        TLDBaseKeys.REGISTRANT_STATE: r"Registrant State/Province: *(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"Registrant Postal Code: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"Registrant Country: *(.+)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Registrant Email: *(.+)",
        TLDBaseKeys.REGISTRANT_PHONE: r"Registrant Phone: *(.+)",
        TLDBaseKeys.REGISTRANT_FAX: r"Registrant Fax: *(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC: *([\S]+)",
        TLDBaseKeys.STATUS: r"Status: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name server: *(.+)",
        TLDBaseKeys.ADMIN_NAME: r"Admin Name: (.+)",
        TLDBaseKeys.ADMIN_ID: r"Admin ID: (.+)",
        TLDBaseKeys.ADMIN_ORGANIZATION: r"Admin Organization: (.+)",
        TLDBaseKeys.ADMIN_CITY: r"Admin City: (.*)",
        TLDBaseKeys.ADMIN_ADDRESS: r"Admin Street: (.*)",
        TLDBaseKeys.ADMIN_STATE: r"Admin State/Province: (.*)",
        TLDBaseKeys.ADMIN_ZIPCODE: r"Admin Postal Code: (.*)",
        TLDBaseKeys.ADMIN_COUNTRY: r"Admin Country: (.+)",
        TLDBaseKeys.ADMIN_PHONE: r"Admin Phone: (.+)",
        TLDBaseKeys.ADMIN_FAX: r"Admin Fax: (.+)",
        TLDBaseKeys.ADMIN_EMAIL: r"Admin Email: (.+)",
        TLDBaseKeys.BILLING_NAME: r"Billing Name: (.+)",
        TLDBaseKeys.BILLING_ID: r"Billing ID: (.+)",
        TLDBaseKeys.BILLING_ORGANIZATION: r"Billing Organization: (.+)",
        TLDBaseKeys.BILLING_CITY: r"Billing City: (.*)",
        TLDBaseKeys.BILLING_ADDRESS: r"Billing Street: (.*)",
        TLDBaseKeys.BILLING_STATE: r"Billing State/Province: (.*)",
        TLDBaseKeys.BILLING_ZIPCODE: r"Billing Postal Code: (.*)",
        TLDBaseKeys.BILLING_COUNTRY: r"Billing Country: (.+)",
        TLDBaseKeys.BILLING_PHONE: r"Billing Phone: (.+)",
        TLDBaseKeys.BILLING_FAX: r"Billing Fax: (.+)",
        TLDBaseKeys.BILLING_EMAIL: r"Billing Email: (.+)",
        TLDBaseKeys.TECH_NAME: r"Tech Name: (.+)",
        TLDBaseKeys.TECH_ID: r"Tech ID: (.+)",
        TLDBaseKeys.TECH_ORGANIZATION: r"Tech Organization: (.+)",
        TLDBaseKeys.TECH_CITY: r"Tech City: (.*)",
        TLDBaseKeys.TECH_ADDRESS: r"Tech Street: (.*)",
        TLDBaseKeys.TECH_STATE: r"Tech State/Province: (.*)",
        TLDBaseKeys.TECH_ZIPCODE: r"Tech Postal Code: (.*)",
        TLDBaseKeys.TECH_COUNTRY: r"Tech Country: (.+)",
        TLDBaseKeys.TECH_PHONE: r"Tech Phone: (.+)",
        TLDBaseKeys.TECH_FAX: r"Tech Fax: (.+)",
        TLDBaseKeys.TECH_EMAIL: r"Tech Email: (.+)",
    }

    multiple_match_keys = (TLDBaseKeys.NAME_SERVERS, TLDBaseKeys.STATUS)
    date_keys = (TLDBaseKeys.CREATED, TLDBaseKeys.UPDATED, TLDBaseKeys.EXPIRES)

    def __init__(self):
        self.reg_expressions = self.base_expressions.copy()


class DomainParser:
    _no_match_checks = [
        "no match",
        "not found",
        "no entries found",
        "invalid query",
        "domain name not known",
        "no object found",
        "available for re-registration",
        "object does not exist",
        "domain you requested is not known",
    ]

    def __init__(self, tld: str):
        self.parser_output = {}
        self._parser = self._init_parser(tld)

    def parse(self, blob: str) -> None:
        """
        Parses `blob` (whois query output) using a parser class.
        Saves the results into the `parser_output` attribute.
        """
        if any([n in blob.lower() for n in self._no_match_checks]):
            raise NotFoundError(f"Domain not found!")
        self.parser_output = self._parser.parse(blob)

    @staticmethod
    def _init_parser(tld: str) -> TLDParser:
        """
        Retrieves the parser instance which can most accurately extract
        key/value pairs from the whois server output for the given `tld`.

        :param tld: the top level domain
        :return: instance of TLDParser or a TLDParser subclass
        """
        if tld == "ae":
            return RegexAE()
        elif tld == "ar":
            return RegexAR()
        elif tld == "at":
            return RegexAT()
        elif tld == "au":
            return RegexAU()
        elif tld == "aw":
            return RegexAW()
        elif tld == "ax":
            return RegexAX()
        elif tld == "be":
            return RegexBE()
        elif tld == "br":
            return RegexBR()
        elif tld == "by":
            return RegexBY()
        elif tld == "cc":
            return RegexCC()
        elif tld == "ch":
            return RegexCH()
        elif tld == "cl":
            return RegexCL()
        elif tld == "cn":
            return RegexCN()
        elif tld == "cr":
            return RegexCR()
        elif tld == "cz":
            return RegexCZ()
        elif tld == "de":
            return RegexDE()
        elif tld == "dk":
            return RegexDK()
        elif tld == "edu":
            return RegexEDU()
        elif tld == "ee":
            return RegexEE()
        elif tld == "eu":
            return RegexEU()
        elif tld == "fi":
            return RegexFI()
        elif tld == "fr":
            return RegexFR()
        elif tld == "ga":
            return RegexGA()
        elif tld == "ge":
            return RegexGE()
        elif tld == "gg":
            return RegexGG()
        elif tld == "gq":
            return RegexGQ()
        elif tld == "hk":
            return RegexHK()
        elif tld == "hr":
            return RegexHR()
        elif tld == "id":
            return RegexID()
        elif tld == "ie":
            return RegexIE()
        elif tld == "il":
            return RegexIL()
        elif tld == "ir":
            return RegexIR()
        elif tld == "is":
            return RegexIS()
        elif tld == "it":
            return RegexIT()
        elif tld == "jp":
            return RegexJP()
        elif tld == "kg":
            return RegexKG()
        elif tld == "kr":
            return RegexKR()
        elif tld == "kz":
            return RegexKZ()
        elif tld == "li":
            return RegexLI()
        elif tld == "lu":
            return RegexLU()
        elif tld == "lv":
            return RegexLV()
        elif tld == "ma":
            return RegexMA()
        elif tld == "ml":
            return RegexML()
        elif tld == "mx":
            return RegexMX()
        elif tld == "nl":
            return RegexNL()
        elif tld == "no":
            return RegexNO()
        elif tld == "nu":
            return RegexNU()
        elif tld == "nz":
            return RegexNZ()
        elif tld == "om":
            return RegexOM()
        elif tld == "pe":
            return RegexPE()
        elif tld == "pl":
            return RegexPL()
        elif tld == "pt":
            return RegexPT()
        elif tld == "rf":
            return RegexRF()
        elif tld == "ro":
            return RegexRO()
        elif tld == "ru":
            return RegexRU()
        elif tld == "sa":
            return RegexSA()
        elif tld == "se":
            return RegexSE()
        elif tld == "si":
            return RegexSI()
        elif tld == "sk":
            return RegexSK()
        elif tld == "su":
            return RegexSU()
        elif tld == "tk":
            return RegexTK()
        elif tld == "tr":
            return RegexTR()
        elif tld == "tw":
            return RegexTW()
        elif tld == "ua":
            return RegexUA()
        elif tld == "uk":
            return RegexUK()
        elif tld == "uz":
            return RegexUZ()
        elif tld == "ve":
            return RegexVE()

        # The TLDParser can handle all "Generic" and some "Country-Code" TLDs.
        # If the parsed output of lookup is not what you expect or even incorrect,
        # check and modify the existing Regex subclass or create a new one.
        return TLDParser()


# ==============================
# Custom Query Output Parsers
# ==============================


class RegexRU(TLDParser):
    _ru_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.CREATED: r"created: *(.+)",
        TLDBaseKeys.EXPIRES: r"paid-till: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org: *(.+)",
        TLDBaseKeys.STATUS: r"state: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.ADMIN_EMAIL: r"admin-contact: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ru_expressions)


class RegexCL(TLDParser):
    _cl_expressions = {
        TLDBaseKeys.NAME_SERVERS: r"Name server: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant name: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Registrant organisation: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar name: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expiration date: (\d{4}-\d{2}-\d{2})",
        TLDBaseKeys.CREATED: r"Creation date: (\d{4}-\d{2}-\d{2})",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._cl_expressions)


class RegexPL(TLDParser):
    _pl_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"DOMAIN NAME: *(.+)\n",
        TLDBaseKeys.REGISTRAR: r"REGISTRAR:\s*(.+)",
        TLDBaseKeys.CREATED: r"created: *(.+)",
        TLDBaseKeys.EXPIRES: r"option expiration date: *(.+)",
        TLDBaseKeys.UPDATED: r"last modified: *(.+)\n",
        TLDBaseKeys.DNSSEC: r"dnssec: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._pl_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        nameservers_match = self.find_match(
            r"nameservers:*(.+)\ncreated:\s", blob, flags=re.DOTALL | re.IGNORECASE
        )
        if nameservers_match:
            parsed_output[TLDBaseKeys.NAME_SERVERS] = [
                self._process(m) for m in nameservers_match.split("\n")
            ]
        return parsed_output


class RegexRO(TLDParser):
    # % The ROTLD WHOIS service on port 43 never discloses any information concerning the registrant.

    _ro_expressions = {
        TLDBaseKeys.CREATED: r"Registered On: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expires On: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Nameserver: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Referral URL: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ro_expressions)


class RegexPE(TLDParser):
    _pe_expressions = {
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant name: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Sponsoring Registrar: *(.+)",
        TLDBaseKeys.ADMIN_EMAIL: r"Admin Email: *(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name server: *(.+)",
        TLDBaseKeys.STATUS: r"Domain Status: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._pe_expressions)


class RegexEE(TLDParser):
    _ee_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *[\n\r]+\s*name: *([^\n\r]+)",
        TLDBaseKeys.STATUS: r"status: *([^\n\r]+)",
        TLDBaseKeys.CREATED: r"registered: *([^\n\r]+)",
        TLDBaseKeys.UPDATED: r"changed: *([^\n\r]+)",
        TLDBaseKeys.EXPIRES: r"expire: *([^\n\r]+)",
        TLDBaseKeys.REGISTRAR: r"(?<=Registrar)[\s\S]*?name: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"(?<=Registrar)[\s\S]*?url: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"(?<=Registrar)[\s\S]*?phone: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"country: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ee_expressions)


class RegexFR(TLDParser):
    _fr_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.CREATED: r"created: (\d{4}-\d{2}-\d{2})",
        TLDBaseKeys.UPDATED: r"last-update: (\d{4}-\d{2}-\d{2})",
        TLDBaseKeys.EXPIRES: r"Expiry Date: (\d{4}-\d{2}-\d{2})",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._fr_expressions)


class RegexBR(TLDParser):
    _br_expressions = {
        TLDBaseKeys.CREATED: r"created: ",
        TLDBaseKeys.UPDATED: r"changed: ",
        TLDBaseKeys.STATUS: r"status: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"responsible: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"country: *(.+)",
        TLDBaseKeys.EXPIRES: r"expires: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._br_expressions)


class RegexKR(TLDParser):
    _kr_expressions = {
        TLDBaseKeys.CREATED: r"Registered Date *: (\d{4}\.\s\d{2}\.\s\d{2}\.)",
        TLDBaseKeys.UPDATED: r"Last Updated Date *: (\d{4}\.\s\d{2}\.\s\d{2}\.)",
        TLDBaseKeys.EXPIRES: r"Expiration Date *: (\d{4}\.\s\d{2}\.\s\d{2}\.)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant *: (.+)",
        TLDBaseKeys.ADMIN_EMAIL: r"Administrative Contact(AC) *: *(.+)",
        TLDBaseKeys.ADMIN_NAME: r"AC E-Mail *: *(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC *: ([a-zA-Z]+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"Registrant Zip Code: *: (.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Registrant Address *: (.+)",
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name *: (.+)",
        TLDBaseKeys.NAME_SERVERS: r"Host Name *: (.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._kr_expressions)


class RegexEU(TLDParser):
    # .EU whois server disclaimer:
    # % The EURid WHOIS service on port 43 (textual whois) never
    # % discloses any information concerning the registrant.

    _eu_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar:\n.*Name: (.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Registrar:\n.*Name:.+\n.*Website: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._eu_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # find name servers
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name servers:", blob
        )
        return parsed_output


class RegexDE(TLDParser):
    # .de disclaimer (very hard to extract information from this provider):
    #
    # % The DENIC whois service on port 43 doesn't disclose any information concerning
    # % the domain holder, general request and abuse contact.
    # % This information can be obtained through use of our web-based whois service
    # % available at the DENIC website:
    # % http://www.denic.de/en/domains/whois-service/web-whois.html

    _de_expressions = {
        TLDBaseKeys.UPDATED: r"Changed: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Nserver: *(.+)",
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._de_expressions)


class RegexUK(TLDParser):
    _uk_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain name:\r\n*(.+)",
        TLDBaseKeys.CREATED: r"Registered on:\s*(\d{2}-\w{3}-\d{4})",
        TLDBaseKeys.UPDATED: r"Last updated:\s*(\d{2}-\w{3}-\d{4})",
        TLDBaseKeys.EXPIRES: r"Expiry date:\s*(\d{2}-\w{3}-\d{4})",
        TLDBaseKeys.REGISTRAR: r"Registrar:\s*(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"(?<=Registrar)[\s\S]*?URL: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant:\n *(.+)",
        TLDBaseKeys.STATUS: r"Registration status:\n *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._uk_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # handle registrant address
        address_match = re.search(
            r"Registrant's address: *(.+)Data valid", blob, re.DOTALL
        )
        if address_match:
            address_pieces = [
                m.strip() for m in address_match.group(1).split("\n") if m.strip()
            ]
            parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = ", ".join(address_pieces)
        # find name servers
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name servers:", blob
        )
        return parsed_output


class RegexJP(TLDParser):
    _jp_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"\[Domain Name\] *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"\[Registrant\] *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"\[Organization\] *(.+)",
        TLDBaseKeys.CREATED: r"\[登録年月日\] *(.+)",
        TLDBaseKeys.EXPIRES: r"\[(?:有効限|有効期限)\]*(.+)",
        TLDBaseKeys.STATUS: r"\[状態\] *(.+)",
        TLDBaseKeys.UPDATED: r"\[最終更新\] *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"\[Name Server\] *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._jp_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # check for domain in japanese
        if not parsed_output.get(TLDBaseKeys.DOMAIN_NAME):
            parsed_output[TLDBaseKeys.DOMAIN_NAME] = self.find_match(
                r"\[ドメイン名\] *(.+)", blob
            )
        # check for name servers in japanese
        if not parsed_output.get(TLDBaseKeys.NAME_SERVERS):
            parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_match(
                r"\[ネームサーバ\] *(.+)", blob, many=True
            )
        # check for address
        address_match = re.search(
            r"\[Postal Address\]([^\[|.]+)\[\w+\](.+)", blob, re.DOTALL
        )
        if address_match:
            address_pieces = [
                m.strip() for m in address_match.group(1).split("\n") if m.strip()
            ]
            parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = ", ".join(address_pieces)
        return parsed_output


class RegexAU(TLDParser):
    _au_expressions = {
        TLDBaseKeys.UPDATED: r"Last Modified: (\d{2}-\w{3}-\d{4})",
        TLDBaseKeys.REGISTRAR: r"Registrar Name:\s *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._au_expressions)


class RegexAT(TLDParser):
    _at_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.UPDATED: r"changed: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"registrant: *(.+)",
        TLDBaseKeys.TECH_NAME: r"tech-c: *(.+)",
    }

    _contact_fields = {
        "name": r"personname: *(.+)",
        "address": r"street address: *(.+)",
        "city": r"city: *(.+)",
        "country": r"country: *(.+)",
        "zipcode": r"postal code: *(.+)",
        "email": "e-mail: *(.+)",
        "phone": "phone: *(.+)",
        "fax": "fax-no: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._at_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parser_output = super().parse(blob)
        contact_field_fills = [
            ["registrant", r"registrant: *(.+)"],
            ["admin", r"admin-c"],
            ["tech", r"tech-c: *(.+)"],
        ]
        # find and save nic-hdls
        for field in contact_field_fills:
            field[1] = self.find_match(field[1], blob) or ""
        # parse contact info using each "nic-hdl"
        for prefix, handle in contact_field_fills:
            pattern = re.compile(
                r"(?:personname):.+\n(?:.+\n){{1,}}nic-hdl: *{nic_hdl}\n(?:.+\n){{1,}}".format(
                    nic_hdl=handle
                ),
                flags=re.I,
            )
            contact_blob = pattern.search(blob)
            if contact_blob:
                for field, field_regex in self._contact_fields.items():
                    key = getattr(TLDBaseKeys, f"{prefix}_{field}".upper())
                    parser_output[key] = self.find_match(
                        field_regex, contact_blob.group()
                    )

        return parser_output


class RegexBE(TLDParser):
    _be_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.CREATED: r"Registered: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar:\n.+Name: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Registrar:\n.+Name:.+\n.+Website:*(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant:\n *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._be_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name servers:", blob
        )
        return parsed_output


class RegexRF(TLDParser):  # same as RU
    def __init__(self):
        super().__init__()

        self.update_reg_expressions(RegexRU._ru_expressions)


class RegexSU(TLDParser):  # same as RU
    def __init__(self):
        super().__init__()
        self.update_reg_expressions(RegexRU._ru_expressions)


class RegexKG(TLDParser):
    _kg_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain *(.+) ",
        TLDBaseKeys.REGISTRAR: r"Domain support: \s*(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Name: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Address: *(.+)",
        TLDBaseKeys.CREATED: r"Record created: *(.+)",
        TLDBaseKeys.EXPIRES: r"Record expires on \s*(.+)",
        TLDBaseKeys.UPDATED: r"Record last updated on\s*(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._kg_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name servers in the listed order:", blob
        )
        return parsed_output


class RegexCH(TLDParser):
    _ch_expressions = {
        TLDBaseKeys.REGISTRANT_NAME: r"Holder of domain name:\s*(?:.*\n){1}\s*(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Holder of domain name:\s*(?:.*\n){2}\s*(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar:\n*(.+)",
        TLDBaseKeys.CREATED: r"First registration date:\n*(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC:*([\S]+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ch_expressions)


class RegexLI(TLDParser):  # same as CH
    def __init__(self):
        super().__init__()
        self.update_reg_expressions(RegexCH._ch_expressions)


class RegexID(TLDParser):
    _id_expressions = {
        TLDBaseKeys.CREATED: r"Created On:(.+)",
        TLDBaseKeys.EXPIRES: r"Expiration Date:(.+)",
        TLDBaseKeys.UPDATED: r"Last Updated On:(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC:(.+)",
        TLDBaseKeys.REGISTRAR: r"Sponsoring Registrar Organization:(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"Sponsoring Registrar Phone: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"Sponsoring Registrar Email: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Sponsoring Registrar URL: *(.+)",
        TLDBaseKeys.STATUS: r"Status:(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Name:(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Registrant Street1:(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._id_expressions)


class RegexSE(TLDParser):
    _se_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"holder\.*: *(.+)",
        TLDBaseKeys.CREATED: r"created\.*: *(.+)",
        TLDBaseKeys.UPDATED: r"modified\.*: *(.+)",
        TLDBaseKeys.EXPIRES: r"expires\.*: *(.+)",
        TLDBaseKeys.DNSSEC: r"dnssec\.*: *(.+)",
        TLDBaseKeys.STATUS: r"status\.*: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._se_expressions)


class RegexIT(TLDParser):
    _it_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.CREATED: r"(?<! )Created: *(.+)",
        TLDBaseKeys.UPDATED: r"(?<! )Last Update: *(.+)",
        TLDBaseKeys.EXPIRES: r"(?<! )Expire Date: *(.+)",
        TLDBaseKeys.STATUS: r"Status: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Registrant)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Registrant)[\s\S]*?Address:(.*)",
        TLDBaseKeys.REGISTRAR: r"(?<=Registrar)[\s\S]*?Name: *(.*)",
        TLDBaseKeys.REGISTRAR_URL: r"(?<=Registrar)[\s\S]*?Web: *(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._it_expressions)


class RegexSA(TLDParser):
    _sa_expressions = {
        TLDBaseKeys.CREATED: r"Created on: *(.+)",
        TLDBaseKeys.UPDATED: r"Last Updated on: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant:\s*(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Registrant)[\s\S]*?Address:((?:.+\n)*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._sa_expressions)


class RegexSK(TLDParser):
    _sk_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.CREATED: r"(?<=Domain:)[\s\w\W]*?Created: *(.+)",
        TLDBaseKeys.UPDATED: r"(?<=Domain:)[\s\w\W]*?Updated: *(.+)",
        TLDBaseKeys.EXPIRES: r"Valid Until: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Name:\s*(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Street:\s*(.+)",
        TLDBaseKeys.REGISTRAR: r"(?<=Registrar)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"(?<=Registrar)[\s\S]*?Email: *(.*)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"(?<=Registrar)[\s\S]*?Phone: *(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=^Contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"(?<=^Contact)[\s\S]*?Postal Code:(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=^Contact)[\s\S]*?Country Code:(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._sk_expressions)


class RegexMX(TLDParser):
    _mx_expressions = {
        TLDBaseKeys.CREATED: r"Created On: *(.+)",
        TLDBaseKeys.UPDATED: r"Last Updated On: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expiration Date: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar:\s*(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"URL: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Registrant)[\s\S]*?Name:(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=Registrant)[\s\S]*?City:(.*)",
        TLDBaseKeys.REGISTRANT_STATE: r"(?<=Registrant)[\s\S]*?State:(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=Registrant)[\s\S]*?Country:(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._mx_expressions)


class RegexTW(TLDParser):
    _tw_expressions = {
        TLDBaseKeys.CREATED: r"Record created on (.+) ",
        TLDBaseKeys.EXPIRES: r"Record expires on (.+) ",
        TLDBaseKeys.REGISTRAR: r"Registration Service Provider: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"Registration Service URL: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Registrant:)\s+(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=Registrant:)\s*(?:.*\n){5}\s+(.*),",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Registrant:)\s*(?:.*\n){4}\s+(.*)",
        TLDBaseKeys.REGISTRANT_STATE: r"(?<=Registrant:)\s*(?:.*\n){5}.*, (.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=Registrant:)\s*(?:.*\n){6}\s+(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._tw_expressions)


class RegexTR(TLDParser):
    _tr_expressions = {
        TLDBaseKeys.CREATED: r"Created on.*: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expires on.*: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=[**] Registrant:)[\s\S]((?:\s.+)*)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=[**] Administrative Contact)[\s\S]*?Address\s+: (.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._tr_expressions)


class RegexIS(TLDParser):
    _is_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.CREATED: r"created\.*: *(.+)",
        TLDBaseKeys.EXPIRES: r"expires\.*: *(.+)",
        TLDBaseKeys.DNSSEC: r"dnssec\.*: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver\.*: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._is_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # find first instance of person/role (registrant)
        registrant_block = False
        registrant_name = None
        addresses = []
        for line in blob.split("\n"):
            if line.startswith("role") or line.startswith("person"):
                if registrant_block:
                    break
                else:
                    registrant_name = line.split(":")[-1].strip()
                    registrant_block = True
            elif line.startswith("address:"):
                address = line.split(":")[-1].strip()
                addresses.append(address)

        parsed_output[TLDBaseKeys.REGISTRANT_NAME] = registrant_name
        # join the address lines together and save
        parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = ", ".join(addresses)
        return parsed_output


class RegexDK(TLDParser):
    _dk_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.CREATED: r"Registered: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expires: *(.+)",
        TLDBaseKeys.DNSSEC: r"Dnssec: *(.+)",
        TLDBaseKeys.STATUS: r"Status: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant\s*(?:.*\n){2}\s*Name: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Registrant\s*(?:.*\n){3}\s*Address: *(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"Registrant\s*(?:.*\n){4}\s*Postalcode: *(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"Registrant\s*(?:.*\n){5}\s*City: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"Registrant\s*(?:.*\n){6}\s*Country: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._dk_expressions)


class RegexIL(TLDParser):
    _li_expressions = {
        TLDBaseKeys.EXPIRES: r"validity: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"person: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"address *(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC: *(.+)",
        TLDBaseKeys.STATUS: r"status: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar name: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._li_expressions)


class RegexFI(TLDParser):
    _fi_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain\.*: *([\S]+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Holder\s*name\.*:\s(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"[Holder\w\W]address\.*: ([\S\ ]+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"[Holder\w\W]address\.*:.+\naddress\.*:\s(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"[Holder\w\W]address\.*:.+\naddress\.*:.+\naddress\.*:\s(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"country\.*:\s(.+)",
        TLDBaseKeys.STATUS: r"status\.*: *([\S]+)",
        TLDBaseKeys.CREATED: r"created\.*: *([\S]+)",
        TLDBaseKeys.UPDATED: r"modified\.*: *([\S]+)",
        TLDBaseKeys.EXPIRES: r"expires\.*: *([\S]+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver\.*: *([\S]+) \[\S+\]",
        TLDBaseKeys.DNSSEC: r"dnssec\.*: *([\S]+)",
        TLDBaseKeys.REGISTRAR: r"registrar\.*:\s(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._fi_expressions)


class RegexNU(TLDParser):
    _nu_expression = {
        TLDBaseKeys.DOMAIN_NAME: r"domain\.*: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"holder\.*: *(.+)",
        TLDBaseKeys.CREATED: r"created\.*: *(.+)",
        TLDBaseKeys.UPDATED: r"modified\.*: *(.+)",
        TLDBaseKeys.EXPIRES: r"expires\.*: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver\.*: *(.+)",
        TLDBaseKeys.DNSSEC: r"dnssec\.*: *(.+)",
        TLDBaseKeys.STATUS: r"status\.*: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._nu_expression)


class RegexPT(TLDParser):
    _pt_expression = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.CREATED: r"Creation Date: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expiration Date: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Owner Name: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Owner Address: *(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"Owner Locality: *(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"Owner ZipCode: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name Server: *(.+) \|",
        TLDBaseKeys.STATUS: r"Domain Status: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._pt_expression)


class RegexIE(TLDParser):
    _ie_expressions = {
        TLDBaseKeys.REGISTRANT_NAME: r"Domain Holder: *(.+)",
        TLDBaseKeys.CREATED: r"Registration Date: *(.+)",
        TLDBaseKeys.EXPIRES: r"Renewal Date: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Nserver: *(.+)",
        TLDBaseKeys.STATUS: r"Renewal status: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Account Name: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ie_expressions)


class RegexNZ(TLDParser):
    # These don't seem to be valid anymore:
    # _nz_expressions = {
    #     TLDBaseKeys.DOMAIN_NAME: r"",
    #     TLDBaseKeys.REGISTRAR: r"registrar_name:\s*([^\n\r]+)",
    #     TLDBaseKeys.UPDATED: r"domain_datelastmodified:\s*([^\n\r]+)",
    #     TLDBaseKeys.CREATED: r"domain_dateregistered:\s*([^\n\r]+)",
    #     TLDBaseKeys.EXPIRES: r"domain_datebilleduntil:\s*([^\n\r]+)",
    #     TLDBaseKeys.NAME_SERVERS: r"ns_name_\d*:\s*([^\n\r]+)",
    #     TLDBaseKeys.STATUS: r"status:\s*([^\n\r]+)",
    #     TLDBaseKeys.REGISTRANT_NAME: r"registrant_contact_name:\s*([^\n\r]+)",
    #     TLDBaseKeys.REGISTRANT_ADDRESS: r"registrant_contact_address\d*:\s*([^\n\r]+)",
    #     TLDBaseKeys.REGISTRANT_CITY: r"registrant_contact_city:\s*([^\n\r]+)",
    #     TLDBaseKeys.REGISTRANT_ZIPCODE: r"registrant_contact_postalcode:\s*([^\n\r]+)",
    #     TLDBaseKeys.REGISTRANT_COUNTRY: r"registrant_contact_country:\s*([^\n\r]+)",
    # }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions({})


class RegexLU(TLDParser):
    _lu_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domainname: *(.+)",
        TLDBaseKeys.CREATED: r"registered: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.STATUS: r"domaintype: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar-name: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"registrar-email: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"registrar-url: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org-name: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"org-address: *(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"org-zipcode:*(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"org-city: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"org-country: *(.+)",
    }

    def __init__(self):
        super().__init__()

        self.update_reg_expressions(self._lu_expressions)


class RegexCZ(TLDParser):
    _cz_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"name: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.CREATED: r"registered: *(.+)",
        TLDBaseKeys.UPDATED: r"changed: *(.+)",
        TLDBaseKeys.EXPIRES: r"expire: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._cz_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        addresses = []
        seen_contact = False
        # extract address from registrant info block
        for line in blob.split("\n"):
            # check that this is first "contact" block
            if line.startswith("contact:"):
                if not seen_contact:
                    seen_contact = True
                else:
                    # if not; stop
                    break
            # append address
            elif line.startswith("address:"):
                line = line.split(":")[-1].strip()
                addresses.append(line)
        # just combine address lines; don't assume city/zipcode/country happen in any specific order
        address = ", ".join(addresses)
        parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = address
        return parsed_output


class RegexHR(TLDParser):
    _hr_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name: *(.+)",
        TLDBaseKeys.UPDATED: r"Updated Date: *(.+)",
        TLDBaseKeys.CREATED: r"Creation Date: *(.+)",
        TLDBaseKeys.EXPIRES: r"Registrar Registration Expiration Date: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name Server: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Name:\s(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Registrant Street:\s*(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._hr_expressions)


class RegexHK(TLDParser):
    _hk_expressions = {
        TLDBaseKeys.STATUS: r"Domain Status: *(.+)",
        TLDBaseKeys.DNSSEC: r"DNSSEC: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar Name: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"Registrar Contact Information: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"Registrar Contact Information: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Contact Information:\s*Company English Name.*:(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Registrant Contact Information:)[\s\S]*?Address: (.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"[Registrant Contact Information\w\W]+Country: ([\S\ ]+)",
        # 'registrant_email': r'[Registrant Contact Information\w\W]+Email: ([\S\ ]+)',
        TLDBaseKeys.UPDATED: r"Updated Date: *(.+)",
        TLDBaseKeys.CREATED: r"[Registrant Contact Information\w\W]+Domain Name Commencement Date: (.+)",
        TLDBaseKeys.EXPIRES: r"[Registrant Contact Information\w\W]+Expiry Date: (.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name Servers Information:\s+((?:.+\n)*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._hk_expressions)


class RegexUA(TLDParser):
    _ua_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.STATUS: r"status: *(.+)",
        TLDBaseKeys.REGISTRAR: r"(?:Registrar: |(?<=Registrar:)[\s\W\w]*?organization-loc: )(.*)",
        TLDBaseKeys.REGISTRAR_URL: r"(?<=Registrar)[\s\S]*?url: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_PHONE: r"(?<=Registrar)[\s\S]*?abuse-phone: *(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"(?<=Registrar)[\s\S]*?abuse-email: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Registrant:)[\s\W\w]*?organization-loc:(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=Registrant:)[\s\W\w]*?country-loc:(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=Registrant:)[\s\W\w]*?(?:address\-loc:\s+.*\n){2}address-loc:\s+(.*)\n",
        TLDBaseKeys.REGISTRANT_STATE: r"(?<=Registrant:)[\s\W\w]*?(?:address\-loc:\s+.*\n){1}address-loc:\s+(.*)\n",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Registrant:)[\s\W\w]*?address-loc:\s+(.*)\n",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"(?<=Registrant:)[\s\W\w]*?postal-code-loc:(.*)",
        TLDBaseKeys.UPDATED: "(?:Updated Date: |modified: )(.+)",
        TLDBaseKeys.CREATED: "(?:Creation Date: |created: )(.+)",
        TLDBaseKeys.EXPIRES: "(?:Registry Expiry Date: |expires: )(.+)",
        TLDBaseKeys.NAME_SERVERS: "nserver: *(.+)",
    }

    KNOWN_DATE_FORMATS = [
        "%Y-%m-%d %H:%M:%S%z",
    ]

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ua_expressions)

    @staticmethod
    def _fix_timezone(date_string: str) -> str:
        """
        Fix timezone format for datetime.strptime

        :param date_string: date string
        :return: fixed date string

        >>> RegexUA._fix_timezone("2023-02-17 14:22:06+02")
        '2023-02-17 14:22:06+0200'
        >>>
        >>> RegexUA._fix_timezone("2023-02-17 14:22:06+2")
        '2023-02-17 14:22:06+0200'
        >>>
        >>> RegexUA._fix_timezone("2023-02-17 14:22:06+0200")
        '2023-02-17 14:22:06+0200'
        >>>

        """
        if "+" in date_string:
            date_time, timezone = date_string.split("+")
            if timezone.isdigit() and len(timezone) <= 2:
                date_string = f"{date_time}+{int(timezone):02d}00"

        return date_string

    @staticmethod
    def _parse_date(date_string: str) -> Union[datetime, str]:
        date = TLDParser._parse_date(date_string)
        if isinstance(date, datetime):
            return date

        date_string = RegexUA._fix_timezone(date_string)
        for date_format in RegexUA.KNOWN_DATE_FORMATS:
            try:
                date = datetime.strptime(date_string, date_format)
                return date
            except ValueError:
                pass

        return date_string


class RegexCN(TLDParser):
    _cn_expressions = {
        TLDBaseKeys.REGISTRAR: r"Sponsoring Registrar: *(.+)",
        TLDBaseKeys.CREATED: r"Registration Time: *(.+)",
        TLDBaseKeys.EXPIRES: r"Expiration Time: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant: *(.+)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Registrant Contact Email: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._cn_expressions)


class RegexAR(TLDParser):
    _ar_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.UPDATED: r"changed: *(.+)",
        TLDBaseKeys.CREATED: r"created: *(.+)",
        TLDBaseKeys.EXPIRES: r"expire: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+) \(.*\)",
        TLDBaseKeys.REGISTRANT_NAME: r"name: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ar_expressions)


class RegexBY(TLDParser):
    _by_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Person: *(.+)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Email: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Org: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"Country: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Address: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._by_expressions)


class RegexCR(TLDParser):
    _cr_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"name: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.UPDATED: r"changed: *(.+)",
        TLDBaseKeys.CREATED: r"registered: *(.+)",
        TLDBaseKeys.EXPIRES: r"expire: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._cr_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        # CR server has the same format as CZ
        return RegexCZ().parse(blob)


class RegexVE(TLDParser):  # double check
    _ve_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
        TLDBaseKeys.CREATED: "registered: *(.+)",
        TLDBaseKeys.EXPIRES: "expire: *(.+)",
        TLDBaseKeys.UPDATED: "changed: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"registrant: *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"address: *(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"address:.+\naddress: *(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"(?:address:.+\n){2}address: *(.+)",
        TLDBaseKeys.REGISTRANT_STATE: r"(?:address:.+\n){3}address: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?:address:.+\n){4}address: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ve_expressions)


class RegexAE(TLDParser):
    _ae_expressions = {
        TLDBaseKeys.STATUS: r"Status: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Contact Name: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Registrant Contact Organisation: *(.+)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Registrant Contact Email: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar Name: *(.+)",
        TLDBaseKeys.TECH_NAME: r"Tech Contact Name: *(.+)",
        TLDBaseKeys.TECH_EMAIL: r"Tech Contact Email: *(.+)",
        TLDBaseKeys.TECH_ORGANIZATION: r"Tech Contact Organisation: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ae_expressions)


class RegexSI(TLDParser):
    _si_expressions = {
        TLDBaseKeys.REGISTRAR: r"registrar: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"registrar-url: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nameserver: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"registrant: *(.+)",
        TLDBaseKeys.CREATED: r"created: *(.+)",
        TLDBaseKeys.EXPIRES: r"expire: *(.+)",
        TLDBaseKeys.DOMAIN_NAME: "domain: *(.+)",
        TLDBaseKeys.STATUS: "status: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._si_expressions)


class RegexNO(TLDParser):
    """
    % The whois service at port 43 is intended to contribute to resolving
    % technical problems where individual domains threaten the
    % functionality, security and stability of other domains or the
    % internet as an infrastructure. It does not give any information
    % about who the holder of a domain is. To find information about a
    % domain holder, please visit our website:
    % https://www.norid.no/en/domeneoppslag/
    """

    _no_expressions = {
        TLDBaseKeys.CREATED: r"Created:\s*(.+)",
        TLDBaseKeys.UPDATED: r"Last updated:\s*(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name Server Handle\.*: *(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar Handle\.*: *(.+)",
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name\.*: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._no_expressions)


class RegexKZ(TLDParser):
    _kz_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain Name.*: (.+)",
        TLDBaseKeys.REGISTRAR: r"Current Registar:\s*(.+)",  # "Registar" typo exists on the whois server
        TLDBaseKeys.CREATED: r"Domain created:\s*(.+)\s\(",
        TLDBaseKeys.UPDATED: r"Last modified\s:\s*(.+)\s\(",
        TLDBaseKeys.NAME_SERVERS: r".+\sserver\.*:\s*(.+)",
        TLDBaseKeys.STATUS: r"Domain status\s:\s(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Organization Using Domain Name\nName\.*:\s(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"Street Address\.*:\s*(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"City\.*:\s*(.+)",
        TLDBaseKeys.REGISTRANT_ZIPCODE: r"Postal Code\.*:\s*(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"Country\.*:\s*(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Organization Name\.*:\s*(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._kz_expressions)


class RegexIR(TLDParser):
    _ir_expressions = {
        TLDBaseKeys.UPDATED: r"last-updated: *(.+)",
        TLDBaseKeys.EXPIRES: r"expire-date: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"org: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"remarks:\s+\(Domain Holder\) *(.+)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"remarks:\s+\(Domain Holder Address\) *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver: *(.+)",
        TLDBaseKeys.DOMAIN_NAME: r"domain: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ir_expressions)


class RegexTK(TLDParser):
    _tk_expressions = {
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"(?<=Owner contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Owner contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Owner contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.REGISTRANT_STATE: r"(?<=Owner contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=Owner contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=Owner contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"(?<=Owner contact)[\s\S]*?E-mail:(.*)",
        TLDBaseKeys.REGISTRANT_FAX: r"(?<=Owner contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.REGISTRANT_PHONE: r"(?<=Owner contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.ADMIN_ORGANIZATION: r"(?<=Admin contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.ADMIN_NAME: r"(?<=Admin contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.ADMIN_ADDRESS: r"(?<=Admin contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.ADMIN_STATE: r"(?<=Admin contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.ADMIN_CITY: r"(?<=Admin contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.ADMIN_COUNTRY: r"(?<=Admin contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.ADMIN_EMAIL: r"(?<=Admin contact)[\s\S]*?E-mail:(.*)",
        TLDBaseKeys.ADMIN_FAX: r"(?<=Admin contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.ADMIN_PHONE: r"(?<=Admin contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.BILLING_ORGANIZATION: r"(?<=Billing contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.BILLING_NAME: r"(?<=Billing contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.BILLING_ADDRESS: r"(?<=Billing contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.BILLING_STATE: r"(?<=Billing contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.BILLING_CITY: r"(?<=Billing contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.BILLING_COUNTRY: r"(?<=Billing contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.BILLING_EMAIL: r"(?<=Billing contact)[\s\S]*?E-mail:(.*)",
        TLDBaseKeys.BILLING_FAX: r"(?<=Billing contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.BILLING_PHONE: r"(?<=Billing contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.TECH_ORGANIZATION: r"(?<=Tech contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.TECH_NAME: r"(?<=Tech contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.TECH_ADDRESS: r"(?<=Tech contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.TECH_STATE: r"(?<=Tech contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.TECH_CITY: r"(?<=Tech contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.TECH_COUNTRY: r"(?<=Tech contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.TECH_EMAIL: r"(?<=Tech contact)[\s\S]*?E-mail:(.*)",
        TLDBaseKeys.TECH_FAX: r"(?<=Tech contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.TECH_PHONE: r"(?<=Tech contact)[\s\S]*?Phone:(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._tk_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # handle multiline nameservers
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain nameservers:", blob
        )
        # a date parser exists for '%d/%m/%Y', but this interferes with the parser needed
        # for this one, which is '%m/%d/%Y', so this date format needs to be parsed separately here
        created_match = re.search(r"Domain registered: *(.+)", blob, re.IGNORECASE)
        if created_match:
            parsed_output[TLDBaseKeys.CREATED] = self._parse_date_mdY(
                created_match.group(1)
            )
        expires_match = re.search(r"Record will expire on: *(.+)", blob, re.IGNORECASE)
        if expires_match:
            parsed_output[TLDBaseKeys.EXPIRES] = self._parse_date_mdY(
                expires_match.group(1)
            )
        # Check if "Status" is inline with Domain Name. For example:
        # Domain Name:
        #   GOOGLE.TK is Active
        #   -- OR --
        #   GOOGLE.TK
        domain_name_match = re.search(r"Domain name:\n*(.+)\n\n", blob, re.IGNORECASE)
        if domain_name_match:
            if " is " in domain_name_match.group(1):
                domain_name, status = domain_name_match.group(1).split(" is ")
                parsed_output[TLDBaseKeys.DOMAIN_NAME] = self._process(
                    domain_name.replace("\n", "")
                )
                parsed_output[TLDBaseKeys.STATUS] = [
                    self._process(status.replace("\n", ""))
                ]
            else:
                parsed_output[TLDBaseKeys.DOMAIN_NAME] = self._process(
                    domain_name_match.group(1).replace("\n", "")
                )
        return parsed_output


class RegexCC(TLDParser):
    _cc_expressions = {TLDBaseKeys.STATUS: r"Domain Status: *(.+)"}

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._cc_expressions)


class RegexEDU(TLDParser):
    _edu_expressions = {
        TLDBaseKeys.CREATED: "Domain record activated: *(.+)",
        TLDBaseKeys.UPDATED: "Domain record last updated: *(.+)",
        TLDBaseKeys.EXPIRES: "Domain expires: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._edu_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        registrant_match = re.search(
            r"Registrant:*(.+)Administrative Contact", blob, re.DOTALL | re.IGNORECASE
        )
        if registrant_match:
            reg_info_raw = registrant_match.group(1).split("\n")
            # remove duplicates and empty strings
            reg_info_clean = []
            for value in reg_info_raw:
                value = value.strip()
                if value and value not in reg_info_clean:
                    reg_info_clean.append(value)
            # country is usually either last or third to last
            country_index = -3
            for i, value in enumerate(reg_info_clean):
                if len(value) == 2:
                    country_index = i
                    break

            org = reg_info_clean[0]
            address = ", ".join(reg_info_clean[1:country_index])
            country = reg_info_clean[country_index]

            parsed_output[TLDBaseKeys.REGISTRANT_NAME] = org
            parsed_output[TLDBaseKeys.REGISTRANT_ORGANIZATION] = org
            parsed_output[TLDBaseKeys.REGISTRANT_COUNTRY] = country
            parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = address

        # handle multiline nameservers
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name Servers:", blob
        )
        return parsed_output


class RegexLV(TLDParser):
    _lv_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain: *(.+)",
        TLDBaseKeys.REGISTRAR: r"\[Registrar\]\n(?:.*)\nName:(.*)+",
        TLDBaseKeys.REGISTRANT_NAME: r"\[Holder\]\n(?:.*)\nName:(.*)+",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"\[Holder\]\n(?:.*)\Address:(.*)+",
        TLDBaseKeys.NAME_SERVERS: r"Nserver: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._lv_expressions)


class RegexGQ(TLDParser):
    _gq_expressions = {}

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._gq_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        # GQ server has the same format as TK
        return RegexTK().parse(blob)


class RegexNL(TLDParser):
    _nl_expressions = {
        TLDBaseKeys.REGISTRAR: r"Registrar:\n(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"Abuse Contact:\n(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._nl_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # handle multiline nameservers
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain nameservers:", blob
        )
        return parsed_output


class RegexMA(TLDParser):
    _ma_expressions = {
        TLDBaseKeys.REGISTRAR: r"Sponsoring Registrar: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ma_expressions)


class RegexGE(TLDParser):
    _ge_expressions = {
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ge_expressions)


class RegexGG(TLDParser):
    _gg_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain:\n*(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant:\n*(.+)",
        TLDBaseKeys.REGISTRAR: r"Registrar:\n*(.+)",
        TLDBaseKeys.CREATED: r"Registered on *(.+) at",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._gg_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        # parse created date
        created_match = parsed_output.get(
            "created"
        )  # looks like 30th April 2003; need to remove day suffix
        if created_match and isinstance(created_match, str):
            date_string = re.sub(r"(\d)(st|nd|rd|th)", r"\1", created_match)
            parsed_output[TLDBaseKeys.CREATED] = datetime.strptime(
                date_string, "%d %B %Y"
            )
        # handle multiline nameservers and statuses
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Name servers:", blob
        )
        parsed_output[TLDBaseKeys.STATUS] = self.find_multiline_match(
            "Domain status:", blob
        )
        return parsed_output


class RegexAW(TLDParser):
    _aw_expressions = {
        TLDBaseKeys.REGISTRAR: r"Registrar:\n*(.+)",
        TLDBaseKeys.REGISTRAR_ABUSE_EMAIL: r"Abuse Contact:\n*(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._aw_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain nameservers:", blob
        )
        return parsed_output


class RegexAX(TLDParser):
    _ax_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"domain\.+: *(.+)",
        TLDBaseKeys.REGISTRAR: r"registrar\.+: *(.+)",
        TLDBaseKeys.REGISTRAR_URL: r"www\.+: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"name\.+: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"country\.+: *(.+)",
        TLDBaseKeys.CREATED: r"created\.+: *(.+)",
        TLDBaseKeys.EXPIRES: r"expires\.+: *(.+)",
        TLDBaseKeys.UPDATED: r"modified\.+: *(.+)",
        TLDBaseKeys.STATUS: r"status\.+: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"nserver\.+: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ax_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        addresses = self.find_match(r"address\.+: *(.+)", blob, many=True)
        if addresses:
            parsed_output[TLDBaseKeys.REGISTRANT_ADDRESS] = ", ".join(addresses)
        return parsed_output


class RegexML(TLDParser):
    _ml_expressions = {
        TLDBaseKeys.EXPIRES: r"Record will expire on: *(.+)",
        TLDBaseKeys.CREATED: r"Domain registered: *(.+)",
        TLDBaseKeys.DOMAIN_NAME: r"Domain name:\n*(.+)\sis\s",
        TLDBaseKeys.STATUS: r"Domain name:\n.+\sis\s*(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ml_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        parsed_output = super().parse(blob)
        parsed_output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain nameservers:", blob
        )
        for contact in ("Admin", "Billing", "Owner", "Tech"):
            # isolate the appropriate contact block
            contact_blob = re.search(f"{contact} contact:\n(.+)\n\n", blob, re.DOTALL)
            if contact_blob:
                if contact == "Owner":
                    # map "owner" to registrant
                    contact = "Registrant"
                for key in (
                    "Organization",
                    "Name",
                    "Address",
                    "Zipcode",
                    "City",
                    "State",
                    "Country",
                    "Phone",
                    "Fax",
                    "E-mail",
                ):
                    # special case: Email -> E-mail
                    if key == "E-mail":
                        base_key = getattr(TLDBaseKeys, f"{contact}_Email".upper())
                    else:
                        base_key = getattr(TLDBaseKeys, f"{contact}_{key}".upper())
                    if not base_key:
                        continue
                    # updated parser dict
                    parsed_output[base_key] = self.find_match(
                        f"{key}: *(.+)", contact_blob.group(0)
                    )
        date_format = "%m/%d/%Y"  # example: 05/28/2013
        if parsed_output.get(TLDBaseKeys.EXPIRES):
            parsed_output[TLDBaseKeys.EXPIRES] = datetime.strptime(
                parsed_output.get(TLDBaseKeys.EXPIRES), date_format
            )
        if parsed_output.get(TLDBaseKeys.CREATED):
            parsed_output[TLDBaseKeys.CREATED] = datetime.strptime(
                parsed_output.get(TLDBaseKeys.CREATED), date_format
            )
        return parsed_output


class RegexOM(TLDParser):
    _om_expressions = {
        TLDBaseKeys.REGISTRAR: r"Registrar Name: *(.+)",
        TLDBaseKeys.UPDATED: r"Last Modified: *(.+)",
        TLDBaseKeys.REGISTRANT_CITY: r"Registrant Contact City: *(.+)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"Registrant Contact Country: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"Registrant Contact Organisation: *(.+)",
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant Contact Name: *(.+)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Registrant Contact Email: *(.+)",
        TLDBaseKeys.TECH_CITY: r"Tech Contact City: *(.+)",
        TLDBaseKeys.TECH_COUNTRY: r"Tech Contact Country: *(.+)",
        TLDBaseKeys.TECH_ORGANIZATION: r"Tech Contact Organisation: *(.+)",
        TLDBaseKeys.TECH_NAME: r"Tech Contact Name: *(.+)",
        TLDBaseKeys.TECH_EMAIL: r"Tech Contact Email: *(.+)",
        TLDBaseKeys.NAME_SERVERS: r"Name Server: *(.+)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._om_expressions)


class RegexUZ(TLDParser):
    _uz_expressions = {
        TLDBaseKeys.REGISTRANT_NAME: r"Registrant:\s+([A-Za-z0-9\.\s]+\n)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"Contact with Registrant: *(.+)",
        TLDBaseKeys.ADMIN_NAME: r"Administrative Contact:\s+([A-Za-z0-9\.\s]+\n)",
        TLDBaseKeys.BILLING_NAME: r"Billing Contact:\s+([A-Za-z0-9\.\s]+\n)",
        TLDBaseKeys.TECH_NAME: r"Technical Contact:\s+([A-Za-z0-9\.\s]+\n)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._uz_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        output = super().parse(blob)
        output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain servers in listed order:", blob
        )
        return output


class RegexGA(TLDParser):
    _ga_expressions = {
        TLDBaseKeys.DOMAIN_NAME: r"Domain name:\n*(.+) is",
        TLDBaseKeys.STATUS: r"Domain name:\n*.+ is (.+)",
        TLDBaseKeys.CREATED: r"Domain registered: *(.+)",
        TLDBaseKeys.EXPIRES: r"Record will expire on: *(.+)",
        TLDBaseKeys.REGISTRANT_ORGANIZATION: r"(?<=Owner contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.REGISTRANT_NAME: r"(?<=Owner contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.REGISTRANT_ADDRESS: r"(?<=Owner contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.REGISTRANT_STATE: r"(?<=Owner contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.REGISTRANT_CITY: r"(?<=Owner contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.REGISTRANT_COUNTRY: r"(?<=Owner contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.REGISTRANT_EMAIL: r"(?<=Owner contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.REGISTRANT_FAX: r"(?<=Owner contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.REGISTRANT_PHONE: r"(?<=Owner contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.ADMIN_ORGANIZATION: r"(?<=Admin contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.ADMIN_NAME: r"(?<=Admin contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.ADMIN_ADDRESS: r"(?<=Admin contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.ADMIN_STATE: r"(?<=Admin contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.ADMIN_CITY: r"(?<=Admin contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.ADMIN_COUNTRY: r"(?<=Admin contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.ADMIN_EMAIL: r"(?<=Admin contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.ADMIN_FAX: r"(?<=Admin contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.ADMIN_PHONE: r"(?<=Admin contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.BILLING_ORGANIZATION: r"(?<=Billing contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.BILLING_NAME: r"(?<=Billing contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.BILLING_ADDRESS: r"(?<=Billing contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.BILLING_STATE: r"(?<=Billing contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.BILLING_CITY: r"(?<=Billing contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.BILLING_COUNTRY: r"(?<=Billing contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.BILLING_EMAIL: r"(?<=Billing contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.BILLING_FAX: r"(?<=Billing contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.BILLING_PHONE: r"(?<=Billing contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.TECH_ORGANIZATION: r"(?<=Tech contact)[\s\S]*?Organization:(.*)",
        TLDBaseKeys.TECH_NAME: r"(?<=Tech contact)[\s\S]*?Name:(.*)",
        TLDBaseKeys.TECH_ADDRESS: r"(?<=Tech contact)[\s\S]*?Address:(.*)",
        TLDBaseKeys.TECH_STATE: r"(?<=Tech contact)[\s\S]*?State:(.*)",
        TLDBaseKeys.TECH_CITY: r"(?<=Tech contact)[\s\S]*?City:(.*)",
        TLDBaseKeys.TECH_COUNTRY: r"(?<=Tech contact)[\s\S]*?Country:(.*)",
        TLDBaseKeys.TECH_EMAIL: r"(?<=Tech contact)[\s\S]*?Phone:(.*)",
        TLDBaseKeys.TECH_FAX: r"(?<=Tech contact)[\s\S]*?Fax:(.*)",
        TLDBaseKeys.TECH_PHONE: r"(?<=Tech contact)[\s\S]*?Phone:(.*)",
    }

    def __init__(self):
        super().__init__()
        self.update_reg_expressions(self._ga_expressions)

    def parse(self, blob: str) -> Dict[str, Any]:
        output = super().parse(blob)
        output[TLDBaseKeys.NAME_SERVERS] = self.find_multiline_match(
            "Domain Nameservers:", blob
        )
        # date format is m/d/Y
        created = output.get(TLDBaseKeys.CREATED)
        if created:
            output[TLDBaseKeys.CREATED] = self._parse_date_mdY(created)

        expires = output.get(TLDBaseKeys.EXPIRES)
        if expires:
            output[TLDBaseKeys.EXPIRES] = self._parse_date_mdY(expires)

        return output



import asyncio
import re
import socket
import sys
from typing import Tuple, Generator
from contextlib import contextmanager

# different installs for async contextmanager based on python version
if sys.version_info < (3, 7):
    from async_generator import asynccontextmanager
else:
    from contextlib import asynccontextmanager

from python_socks.sync import Proxy
from python_socks.async_.asyncio import Proxy as AsyncProxy

BLOCKSIZE = 1024


class Query:
    iana_server = "whois.iana.org"
    whois_port = 43
    refer_regex = r"refer: *(.+)"
    whois_server_regex = r".+ whois server: *(.+)"

    def __init__(
        self,
        server: str = None,
        authoritative_only: bool = True,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        self.server = server
        self.authoritative_only = authoritative_only
        self.proxy_url = proxy_url
        self.timeout = timeout
        self.authoritative_server = ""
        self.query_output = ""
        self.query_chain = ""

    @staticmethod
    def _find_match(regex: str, blob: str) -> str:
        match = ""
        found = re.search(regex, blob, flags=re.IGNORECASE)
        if found:
            match = found.group(1).rstrip("\r").replace(" ", "").rstrip(":").rstrip("/")
        return match

    @contextmanager
    def _create_connection(
        self, address: Tuple[str, int], proxy_url: str = None
    ) -> Generator[socket.socket, None, None]:
        s = None
        try:
            # Use proxy if specified
            if proxy_url:
                proxy = Proxy.from_url(proxy_url)
                # proxy is a standard python socket in blocking mode
                s = proxy.connect(*address, timeout=self.timeout)
            else:
                # otherwise use socket
                s = socket.create_connection(address, self.timeout)
            yield s
        finally:
            if s and hasattr(s, "close"):
                s.close()

    @asynccontextmanager
    async def _aio_create_connection(
        self, address: Tuple[str, int], proxy_url: str = None
    ) -> Generator[Tuple[asyncio.StreamReader, asyncio.StreamWriter], None, None]:
        # init
        reader, writer = None, None
        # Use proxy if specified
        if proxy_url:
            proxy = AsyncProxy.from_url(proxy_url)
            # sock is a standard python socket in blocking mode
            sock = await proxy.connect(*address, timeout=self.timeout)
            # pass it to asyncio
            s = asyncio.open_connection(host=None, port=None, sock=sock)
        else:
            # otherwise use asyncio to open the connection
            s = asyncio.open_connection(*address)
        try:
            reader, writer = await asyncio.wait_for(s, self.timeout)
            yield reader, writer
        finally:
            if writer:
                writer.close()
                if hasattr(writer, "wait_closed"):
                    await writer.wait_closed()

    @staticmethod
    def _send_and_recv(conn: socket.socket, data: str) -> str:
        conn.sendall(data.encode())
        result = ""
        while True:
            received = conn.recv(BLOCKSIZE)
            if received == b"":
                break
            else:
                result += received.decode("utf-8", errors="ignore")
        return result

    @staticmethod
    async def _aio_send_and_recv(
        reader: asyncio.StreamReader, writer: asyncio.StreamWriter, data: str
    ) -> str:
        writer.write(data.encode())
        result = ""
        while True:
            received = await reader.read(BLOCKSIZE)
            if received == b"":
                break
            else:
                result += received.decode("utf-8", errors="ignore")
        return result

    def _do_query(self, server: str, data: str, regex: str) -> str:
        """
        Recursively submits WHOIS queries until it reaches the Authoritative Server.
        Additionally, if `authoritative_only` is False, all text output from server hops
        is saved into `query_chain`.
        """
        # save authoritative server
        self.authoritative_server = server
        # connect to whois://<server>:43
        with self._create_connection((server, self.whois_port), self.proxy_url) as conn:
            # submit domain and receive raw query output
            query_output = self._send_and_recv(conn, data)
            if not self.authoritative_only:
                # concatenate query outputs
                self.query_chain += query_output
            # parse response for the referred WHOIS server name
            whois_server = self._find_match(regex, query_output)
            whois_server = whois_server.lower()
            if (
                whois_server
                and whois_server != server
                and not whois_server.startswith("http")
            ):
                # recursive call to find more authoritative server
                authoritative_output = self._do_query(
                    whois_server, data, self.whois_server_regex
                )
                # check for empty responses; only update query_output if
                # there was a complete response from the authoritative server
                if authoritative_output:
                    query_output = authoritative_output
        # return the WHOIS query output
        return query_output

    async def _aio_do_query(self, server: str, data: str, regex: str):
        # connect to whois://<server>:43
        async with self._aio_create_connection(
            (server, self.whois_port), self.proxy_url
        ) as r_and_w:
            # socket reader and writer
            reader, writer = r_and_w
            # submit domain and receive raw query output
            query_output = await asyncio.wait_for(
                self._aio_send_and_recv(reader, writer, data), self.timeout
            )
            if not self.authoritative_only:
                # concatenate query outputs
                self.query_chain += query_output
            # parse response for the referred WHOIS server name
            whois_server = self._find_match(regex, query_output)
            whois_server = whois_server.lower()
            if (
                whois_server
                and whois_server != server
                and not whois_server.startswith("http")
            ):
                # recursive call to find the authoritative server
                authoritative_output = await self._aio_do_query(
                    whois_server, data, self.whois_server_regex
                )
                # check for empty responses; only update query_output if
                # there was a complete response from the authoritative server
                if authoritative_output:
                    query_output = authoritative_output
        # return the WHOIS query output
        return query_output


class DomainQuery(Query):
    def __init__(
        self,
        domain: str,
        server: str = None,
        authoritative_only: bool = True,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        super().__init__(server, authoritative_only, proxy_url, timeout)
        self.domain = domain

    @classmethod
    def new(
        cls,
        domain: str,
        server: str = None,
        authoritative_only: bool = True,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        _self = cls(domain, server, authoritative_only, proxy_url, timeout)
        data = domain + "\r\n"
        if not _self.server:
            server_regex = _self.refer_regex
            server = _self.iana_server
        else:
            server_regex = _self.whois_server_regex
            server = _self.server
        _self.query_output = _self._do_query(server, data, server_regex)
        return _self

    @classmethod
    async def new_aio(
        cls,
        domain: str,
        server: str = None,
        authoritative_only: bool = True,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        _self = cls(domain, server, authoritative_only, proxy_url, timeout)
        data = domain + "\r\n"
        if not _self.server:
            server_regex = _self.refer_regex
            server = _self.iana_server
        else:
            server_regex = _self.whois_server_regex
            server = _self.server
        _self.query_output = await _self._aio_do_query(server, data, server_regex)
        return _self


class NumberQuery(Query):
    def __init__(
        self,
        ip: str,  # ipv4 or ipv6
        server: str = None,
        authoritative_only: bool = True,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        super().__init__(server, authoritative_only, proxy_url, timeout)
        self.whois_server_regex = r"ReferralServer: *whois://(.+)"
        self.ip = ip

    @classmethod
    def new(
        cls,
        ip: str,  # ipv4 or ipv6; validation is handled by caller.
        server: str = None,
        authoritative_only: bool = False,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        _self = cls(ip, server, authoritative_only, proxy_url, timeout)
        data = ip + "\r\n"
        if ":" in ip:  # ipv6
            _self.refer_regex = r"whois: *(.+)"
        server = server or _self.iana_server
        _self.query_output = _self._do_query(server, data, _self.refer_regex)
        return _self

    @classmethod
    async def new_aio(
        cls,
        ip: str,
        server: str = None,
        authoritative_only: bool = False,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        _self = cls(ip, server, authoritative_only, proxy_url, timeout)
        data = ip + "\r\n"
        if ":" in ip:  # ipv6
            _self.refer_regex = r"whois: *(.+)"
        server = server or _self.iana_server
        _self.query_output = await _self._aio_do_query(server, data, _self.refer_regex)
        return _self



import unittest

from asyncwhois.pywhois import DomainLookup


class TestDomainLookup(unittest.TestCase):

    def setUp(self) -> None:
        self.fake_domain = "some-domain-somewhere-out-there.co.uk"
        self.fake_query_result = "Domain Name: some-domain-somewhere-out-there.co.uk"

    def test__get_server_name(self):
        pyw = DomainLookup()
        generic_tld = 'com'
        whois_server = pyw._get_server_name(generic_tld)
        self.assertEqual(whois_server, 'whois.verisign-grs.com')
        generic_tld_unicode = 'xn--4gbrim'
        whois_server = pyw._get_server_name(generic_tld_unicode)
        self.assertEqual(whois_server, 'whois.afilias-srs.net')
        country_tld = 'us'
        whois_server = pyw._get_server_name(country_tld)
        self.assertEqual(whois_server, 'whois.nic.us')
        sponsored_tld = 'aero'
        whois_server = pyw._get_server_name(sponsored_tld)
        self.assertEqual(whois_server, 'whois.aero')

    def test__get_top_level_domain(self):
        pyw = DomainLookup()
        assert pyw._get_top_level_domain('https://www.google.co.uk') == 'uk'
        assert pyw._get_top_level_domain('https://www.wikipedia.org') == 'org'
        assert pyw._get_top_level_domain('https://www.coral.ai') == 'ai'



import setuptools
import pathlib

base_dir = pathlib.Path(__file__).parent.resolve()


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version(location: str) -> str:
    with open((base_dir / location).absolute().resolve()) as file:
        for line in file.readlines():
            if line.startswith("__version__"):
                return line.split(" = ")[-1].strip().replace('"', "")
        else:
            raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="asyncwhois",
    version=get_version("asyncwhois/__init__.py"),
    author="Joseph Obarzanek",
    author_email="pogzyb@umich.edu",
    description="Python utility for querying and parsing WHOIS information for Domains, IPv4s, IPv6s, and AS numbers.",
    long_description=long_description,
    license="MIT",
    install_requires=[
        "python-socks[asyncio]>=2.0.2",
        "tldextract>=2.2.0",
        "whodap>=0.1.4",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development",
        "Topic :: Security",
        "Framework :: AsyncIO",
    ],
    url="https://github.com/pogzyb/asyncwhois",
    packages=["asyncwhois"],
    package_dir={"asyncwhois": "asyncwhois"},
    python_requires=">=3.6",
)



import asyncio
import sys
import unittest.mock as mock

import asyncwhois
from asyncwhois.servers import CountryCodeTLD
import pytest


test_domain_name = 'amazon.com'
mock_query_data = {
    'parser_output': {'domain_name': test_domain_name},
    'query_output': 'Domain Name: amazon.com'
}


if sys.version_info < (3, 8):
    @pytest.fixture()
    def mock_aio_whois_domain(mocker):
        future = asyncio.Future()
        future.set_result(mock.Mock(
            query_output=mock_query_data.get('query_output'),
            parser_output=mock_query_data.get('parser_output')
        ))
        mocker.patch('asyncwhois.pywhois.DomainLookup.aio_whois_domain', return_value=future)
        return future
else:
    @pytest.fixture()
    def mock_aio_whois_domain(mocker):
        async_mock = mock.AsyncMock(
            return_value=mock.Mock(
                query_output=mock_query_data.get('query_output'),
                parser_output=mock_query_data.get('parser_output')
            )
        )
        mocker.patch('asyncwhois.pywhois.DomainLookup.aio_whois_domain', side_effect=async_mock)
        return async_mock


@pytest.fixture()
def mock_whois_domain(mocker):
    mocker.patch(
        'asyncwhois.pywhois.DomainLookup.whois_domain',
        return_value=mock.Mock(
            query_output=mock_query_data.get('query_output'),
            parser_output=mock_query_data.get('parser_output')
        )
     )


@pytest.mark.asyncio
async def test_aio_lookup(mock_aio_whois_domain):
    result = await asyncwhois.aio_whois_domain(test_domain_name)
    assert f"domain name: {test_domain_name}" in result.query_output.lower(), \
        f"domain name: {test_domain_name} not in {result.query_output.lower()}"
    assert result.parser_output.get('domain_name').lower() == test_domain_name


def test_lookup(mock_whois_domain):
    result = asyncwhois.whois_domain(test_domain_name)
    assert f"domain name: {test_domain_name}" in result.query_output.lower(), \
        f"domain name: {test_domain_name} not in {result.query_output.lower()}"
    assert result.parser_output.get('domain_name').lower() == test_domain_name


def test_input_parameters_for_domain_query(mocker):
    spy = mocker.spy(asyncwhois.query.DomainQuery, 'new')
    mocker.patch(
        'asyncwhois.query.DomainQuery._do_query',
        return_value=mock_query_data.get('query_output')
    )

    test_subdomain_name = 'example.org.ua'

    _ = asyncwhois.whois_domain(test_subdomain_name)
    call_args = spy.call_args_list[0][0]
    assert call_args[0] == test_subdomain_name
    assert call_args[1] == CountryCodeTLD.UA
